import { LocalNode, cojsonInternals } from "cojson";
import { RawAccount, } from "cojson";
import { CoMap, AnonymousJazzAgent, } from "../internal.js";
import { Group, CoValueBase, MembersSym, Profile, Ref, SchemaInit, inspect, subscriptionsScopes, loadCoValue, subscribeToCoValue, ensureCoValueLoaded, subscribeToExistingCoValue, } from "../internal.js";
/** @category Identity & Permissions */
export class Account extends CoValueBase {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    static _schema;
    get _schema() {
        return this.constructor._schema;
    }
    static {
        this._schema = {
            profile: {
                ref: () => Profile,
                optional: false,
            },
            root: {
                ref: () => CoMap,
                optional: true,
            },
        };
    }
    get _owner() {
        return this;
    }
    get _loadedAs() {
        if (this.isMe)
            return this;
        if (this._raw.core.node.account instanceof RawAccount) {
            return Account.fromRaw(this._raw.core.node.account);
        }
        return new AnonymousJazzAgent(this._raw.core.node);
    }
    get _refs() {
        const profileID = this._raw.get("profile");
        const rootID = this._raw.get("root");
        return {
            profile: profileID &&
                new Ref(profileID, this._loadedAs, this._schema.profile),
            root: rootID &&
                new Ref(rootID, this._loadedAs, this._schema.root),
        };
    }
    isMe;
    sessionID;
    constructor(options) {
        super();
        if (!("fromRaw" in options)) {
            throw new Error("Can only construct account from raw or with .create()");
        }
        this.isMe = options.fromRaw.id == options.fromRaw.core.node.account.id;
        Object.defineProperties(this, {
            id: {
                value: options.fromRaw.id,
                enumerable: false,
            },
            _raw: { value: options.fromRaw, enumerable: false },
            _type: { value: "Account", enumerable: false },
        });
        if (this.isMe) {
            this.sessionID = options.fromRaw.core.node.currentSessionID;
        }
        return new Proxy(this, AccountAndGroupProxyHandler);
    }
    myRole() {
        if (this.isMe) {
            return "admin";
        }
    }
    async acceptInvite(valueID, inviteSecret, coValueClass) {
        if (!this.isMe) {
            throw new Error("Only a controlled account can accept invites");
        }
        await this._raw.acceptInvite(valueID, inviteSecret);
        return loadCoValue(coValueClass, valueID, this, []);
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
    }
    /** @private */
    static async create(options) {
        const { node } = await LocalNode.withNewlyCreatedAccount({
            ...options,
            migration: async (rawAccount, _node, creationProps) => {
                const account = new this({
                    fromRaw: rawAccount,
                });
                await account.migrate?.(creationProps);
            },
        });
        return this.fromNode(node);
    }
    static async createAs(as, options) {
        // TODO: is there a cleaner way to do this?
        const connectedPeers = cojsonInternals.connectedPeers("creatingAccount", "createdAccount", { peer1role: "server", peer2role: "client" });
        as._raw.core.node.syncManager.addPeer(connectedPeers[1]);
        return this.create({
            creationProps: options.creationProps,
            crypto: as._raw.core.node.crypto,
            peersToLoadFrom: [connectedPeers[0]],
        });
    }
    static fromNode(node) {
        return new this({
            fromRaw: node.account,
        });
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    toJSON() {
        return {
            id: this.id,
            _type: this._type,
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    migrate(creationProps) {
        if (creationProps) {
            const profileGroup = Group.create({ owner: this });
            profileGroup.addMember("everyone", "reader");
            this.profile = Profile.create({ name: creationProps.name }, { owner: profileGroup });
        }
    }
    /** @category Subscription & Loading */
    static load(id, as, depth) {
        return loadCoValue(this, id, as, depth);
    }
    /** @category Subscription & Loading */
    static subscribe(id, as, depth, listener) {
        return subscribeToCoValue(this, id, as, depth, listener);
    }
    /** @category Subscription & Loading */
    ensureLoaded(depth) {
        return ensureCoValueLoaded(this, depth);
    }
    /** @category Subscription & Loading */
    subscribe(depth, listener) {
        return subscribeToExistingCoValue(this, depth, listener);
    }
}
export const AccountAndGroupProxyHandler = {
    get(target, key, receiver) {
        if (key === "profile") {
            const ref = target._refs.profile;
            return ref
                ? ref.accessFrom(receiver, "profile")
                : // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    undefined;
        }
        else if (key === "root") {
            const ref = target._refs.root;
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            return ref ? ref.accessFrom(receiver, "root") : undefined;
        }
        else {
            return Reflect.get(target, key, receiver);
        }
    },
    set(target, key, value, receiver) {
        if ((key === "profile" || key === "root" || key === MembersSym) &&
            typeof value === "object" &&
            SchemaInit in value) {
            target.constructor._schema ||= {};
            target.constructor._schema[key] =
                value[SchemaInit];
            return true;
        }
        else if (key === "profile") {
            if (value) {
                target._raw.set("profile", value.id, "trusting");
            }
            subscriptionsScopes
                .get(receiver)
                ?.onRefAccessedOrSet(target.id, value.id);
            return true;
        }
        else if (key === "root") {
            if (value) {
                target._raw.set("root", value.id);
            }
            subscriptionsScopes
                .get(receiver)
                ?.onRefAccessedOrSet(target.id, value.id);
            return true;
        }
        else {
            return Reflect.set(target, key, value, receiver);
        }
    },
    defineProperty(target, key, descriptor) {
        if ((key === "profile" || key === "root" || key === MembersSym) &&
            typeof descriptor.value === "object" &&
            SchemaInit in descriptor.value) {
            target.constructor._schema ||= {};
            target.constructor._schema[key] =
                descriptor.value[SchemaInit];
            return true;
        }
        else {
            return Reflect.defineProperty(target, key, descriptor);
        }
    },
};
/** @category Identity & Permissions */
export function isControlledAccount(account) {
    return account.isMe;
}
//# sourceMappingURL=account.js.map