import { RawAccount } from "cojson";
import { Account, Group, SubscriptionScope, Ref, inspect, subscriptionsScopes, AnonymousJazzAgent, } from "../internal.js";
import { fulfillsDepth } from "./deepLoading.js";
// eslint-disable-next-line @typescript-eslint/no-explicit-any
export function isCoValue(value) {
    return value && value._type !== undefined;
}
export function isCoValueClass(
// eslint-disable-next-line @typescript-eslint/no-explicit-any
value) {
    return typeof value === "function" && value.fromRaw !== undefined;
}
/** @internal */
export class CoValueBase {
    get _owner() {
        const owner = this._raw.group instanceof RawAccount
            ? Account.fromRaw(this._raw.group)
            : Group.fromRaw(this._raw.group);
        const subScope = subscriptionsScopes.get(this);
        if (subScope) {
            subScope.onRefAccessedOrSet(this.id, owner.id);
            subscriptionsScopes.set(owner, subScope);
        }
        return owner;
    }
    /** @private */
    get _loadedAs() {
        if (this._raw.core.node.account instanceof RawAccount) {
            return Account.fromRaw(this._raw.core.node.account);
        }
        return new AnonymousJazzAgent(this._raw.core.node);
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    constructor(..._args) {
        Object.defineProperty(this, "_instanceID", {
            value: `instance-${Math.random().toString(36).slice(2)}`,
            enumerable: false,
        });
    }
    /** @category Internals */
    static fromRaw(raw) {
        return new this({ fromRaw: raw });
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    toJSON() {
        return {
            id: this.id,
            type: this._type,
            error: "unknown CoValue class",
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    /** @category Type Helpers */
    castAs(cl) {
        const casted = cl.fromRaw(this._raw);
        const subscriptionScope = subscriptionsScopes.get(this);
        if (subscriptionScope) {
            subscriptionsScopes.set(casted, subscriptionScope);
        }
        return casted;
    }
}
export function loadCoValue(cls, id, as, depth) {
    return new Promise((resolve) => {
        const unsubscribe = subscribeToCoValue(cls, id, as, depth, (value) => {
            resolve(value);
            unsubscribe();
        }, () => {
            resolve(undefined);
            unsubscribe();
        });
    });
}
export function ensureCoValueLoaded(existing, depth) {
    return loadCoValue(existing.constructor, existing.id, existing._loadedAs, depth);
}
export function subscribeToCoValue(cls, id, as, depth, listener, onUnavailable) {
    const ref = new Ref(id, as, { ref: cls, optional: false });
    let unsubscribed = false;
    let unsubscribe;
    ref.load()
        .then((value) => {
        if (!value) {
            onUnavailable && onUnavailable();
            return;
        }
        if (unsubscribed)
            return;
        const subscription = new SubscriptionScope(value, cls, (update) => {
            if (fulfillsDepth(depth, update)) {
                listener(update);
            }
        });
        unsubscribe = () => subscription.unsubscribeAll();
    })
        .catch((e) => {
        console.error("Failed to load / subscribe to CoValue", e);
    });
    return function unsubscribeAtAnyPoint() {
        unsubscribed = true;
        unsubscribe && unsubscribe();
    };
}
export function subscribeToExistingCoValue(existing, depth, listener) {
    return subscribeToCoValue(existing.constructor, existing.id, existing._loadedAs, depth, listener);
}
//# sourceMappingURL=interfaces.js.map