import { ControlledAgent, LocalNode, } from "cojson";
import { Account } from "../internal.js";
export const fixedCredentialsAuth = (credentials) => {
    return {
        start: async () => ({
            type: "existing",
            credentials,
            onSuccess: () => { },
            onError: () => { },
        }),
    };
};
export async function randomSessionProvider(accountID, crypto) {
    return {
        sessionID: crypto.newRandomSessionID(accountID),
        sessionDone: () => { },
    };
}
export async function createJazzContext(options) {
    // eslint-disable-next-line no-constant-condition
    while (true) {
        if (!("auth" in options)) {
            return createAnonymousJazzContext({
                peersToLoadFrom: options.peersToLoadFrom,
                crypto: options.crypto,
            });
        }
        const { auth, sessionProvider, peersToLoadFrom, crypto } = options;
        const AccountSchema = options.AccountSchema ?? Account;
        const authResult = await auth.start(crypto);
        if (authResult.type === "existing") {
            try {
                const { sessionID, sessionDone } = await sessionProvider(authResult.credentials.accountID, crypto);
                try {
                    const node = await LocalNode.withLoadedAccount({
                        accountID: authResult.credentials
                            .accountID,
                        accountSecret: authResult.credentials.secret,
                        sessionID: sessionID,
                        peersToLoadFrom: peersToLoadFrom,
                        crypto: crypto,
                        migration: async (rawAccount, _node, creationProps) => {
                            const account = new AccountSchema({
                                fromRaw: rawAccount,
                            });
                            await account.migrate?.(creationProps);
                        },
                    });
                    const account = AccountSchema.fromNode(node);
                    authResult.onSuccess();
                    return {
                        account,
                        done: () => {
                            node.gracefulShutdown();
                            sessionDone();
                        },
                    };
                }
                catch (e) {
                    authResult.onError(new Error("Error loading account", { cause: e }));
                    sessionDone();
                }
            }
            catch (e) {
                authResult.onError(new Error("Error acquiring sessionID", { cause: e }));
            }
        }
        else if (authResult.type === "new") {
            try {
                // TODO: figure out a way to not "waste" the first SessionID
                const { node } = await LocalNode.withNewlyCreatedAccount({
                    creationProps: authResult.creationProps,
                    peersToLoadFrom: peersToLoadFrom,
                    crypto: crypto,
                    initialAgentSecret: authResult.initialSecret,
                    migration: async (rawAccount, _node, creationProps) => {
                        const account = new AccountSchema({
                            fromRaw: rawAccount,
                        });
                        await account.migrate?.(creationProps);
                    },
                });
                const account = AccountSchema.fromNode(node);
                await authResult.saveCredentials({
                    accountID: node.account.id,
                    secret: node.account.agentSecret,
                });
                authResult.onSuccess();
                return {
                    account,
                    done: () => {
                        node.gracefulShutdown();
                    },
                };
            }
            catch (e) {
                authResult.onError(new Error("Error creating account", { cause: e }));
            }
        }
    }
}
export class AnonymousJazzAgent {
    node;
    _type = "Anonymous";
    constructor(node) {
        this.node = node;
    }
}
export async function createAnonymousJazzContext({ peersToLoadFrom, crypto, }) {
    const agentSecret = crypto.newRandomAgentSecret();
    const rawAgent = new ControlledAgent(agentSecret, crypto);
    const node = new LocalNode(rawAgent, crypto.newRandomSessionID(rawAgent.id), crypto);
    for (const peer of peersToLoadFrom) {
        node.syncManager.addPeer(peer);
    }
    return {
        agent: new AnonymousJazzAgent(node),
        done: () => { },
    };
}
//# sourceMappingURL=createContext.js.map