import { instantiateRefEncoded, isRefEncoded, subscriptionsScopes, } from "../internal.js";
const refCache = new WeakMap();
const TRACE_ACCESSES = false;
export class Ref {
    id;
    controlledAccount;
    schema;
    constructor(id, controlledAccount, schema) {
        this.id = id;
        this.controlledAccount = controlledAccount;
        this.schema = schema;
        if (!isRefEncoded(schema)) {
            throw new Error("Ref must be constructed with a ref schema");
        }
    }
    get value() {
        const node = "node" in this.controlledAccount
            ? this.controlledAccount.node
            : this.controlledAccount._raw.core.node;
        const raw = node.getLoaded(this.id);
        if (raw) {
            let value = refCache.get(raw);
            if (value) {
                // console.log("Using cached value for " + this.id);
            }
            else {
                value = instantiateRefEncoded(this.schema, raw);
                refCache.set(raw, value);
            }
            return value;
        }
        else {
            return null;
        }
    }
    async loadHelper(options) {
        const node = "node" in this.controlledAccount
            ? this.controlledAccount.node
            : this.controlledAccount._raw.core.node;
        const raw = await node.load(this.id, options?.onProgress);
        if (raw === "unavailable") {
            return "unavailable";
        }
        else {
            return new Ref(this.id, this.controlledAccount, this.schema).value;
        }
    }
    async load(options) {
        const result = await this.loadHelper(options);
        if (result === "unavailable") {
            return undefined;
        }
        else {
            return result;
        }
    }
    accessFrom(fromScopeValue, key) {
        const subScope = subscriptionsScopes.get(fromScopeValue);
        subScope?.onRefAccessedOrSet(fromScopeValue.id, this.id);
        TRACE_ACCESSES &&
            console.log(subScope?.scopeID, "accessing", fromScopeValue, key, this.id);
        if (this.value && subScope) {
            subscriptionsScopes.set(this.value, subScope);
        }
        if (subScope) {
            const cached = subScope.cachedValues[this.id];
            if (cached) {
                TRACE_ACCESSES && console.log("cached", cached);
                return cached;
            }
            else if (this.value !== null) {
                const freshValueInstance = instantiateRefEncoded(this.schema, this.value?._raw);
                TRACE_ACCESSES &&
                    console.log("freshValueInstance", freshValueInstance);
                subScope.cachedValues[this.id] = freshValueInstance;
                subscriptionsScopes.set(freshValueInstance, subScope);
                return freshValueInstance;
            }
            else {
                return null;
            }
        }
        else {
            return this.value;
        }
    }
}
export function makeRefs(getIdForKey, getKeysWithIds, controlledAccount, refSchemaForKey) {
    const refs = {};
    return new Proxy(refs, {
        get(_target, key) {
            if (key === Symbol.iterator) {
                return function* () {
                    for (const key of getKeysWithIds()) {
                        yield new Ref(getIdForKey(key), controlledAccount, refSchemaForKey(key));
                    }
                };
            }
            if (typeof key === "symbol")
                return undefined;
            if (key === "length") {
                return getKeysWithIds().length;
            }
            const id = getIdForKey(key);
            if (!id)
                return undefined;
            return new Ref(id, controlledAccount, refSchemaForKey(key));
        },
        ownKeys() {
            return getKeysWithIds().map((key) => key.toString());
        },
        getOwnPropertyDescriptor(target, key) {
            const id = getIdForKey(key);
            if (id) {
                return {
                    enumerable: true,
                    configurable: true,
                    writable: true,
                };
            }
            else {
                return Reflect.getOwnPropertyDescriptor(target, key);
            }
        },
    });
}
//# sourceMappingURL=refs.js.map